import { AudioPlayerStatus, createAudioPlayer, createAudioResource, entersState, getVoiceConnection, joinVoiceChannel, StreamType, VoiceConnectionStatus, } from "@discordjs/voice";
import { StageChannel } from "discord.js";
import chalk from "chalk";
import { getAllAudioUrls } from "google-tts-api";
import ytdl from "ytdl-core";
import { sLogger } from "../logger";
import { shorten } from "../utils/string";
export var VoiceValidateResult;
(function (VoiceValidateResult) {
    VoiceValidateResult[VoiceValidateResult["OK"] = 0] = "OK";
    VoiceValidateResult[VoiceValidateResult["NO_CHANNEL"] = 1] = "NO_CHANNEL";
    VoiceValidateResult[VoiceValidateResult["STAGE_CHANNEL"] = 2] = "STAGE_CHANNEL";
    VoiceValidateResult[VoiceValidateResult["NOT_JOINABLE"] = 3] = "NOT_JOINABLE";
})(VoiceValidateResult || (VoiceValidateResult = {}));
export class VoiceControl {
    constructor(msg, ignorePrivacy, handleDisconnect) {
        this.songRunning = false;
        this.speakQueue = [];
        this.handleDisconnect = handleDisconnect;
        const validateResult = VoiceControl.validateUser(msg);
        if (validateResult)
            throw validateResult;
        this.player = createAudioPlayer();
        const voiceChannel = msg.member.voice.channel;
        this.guildID = voiceChannel.guild.id;
        this.channelName = voiceChannel.name;
        this.voiceChannelID = voiceChannel.id;
        const connection = joinVoiceChannel({
            channelId: voiceChannel.id,
            guildId: voiceChannel.guild.id,
            adapterCreator: voiceChannel.guild
                .voiceAdapterCreator,
            selfMute: ignorePrivacy,
        });
        connection.subscribe(this.player);
        // * https://discordjs.guide/voice/voice-connections.html#handling-disconnects
        connection.on(VoiceConnectionStatus.Disconnected, async (oldState, newState) => {
            try {
                await Promise.race([
                    entersState(connection, VoiceConnectionStatus.Signalling, 5000),
                    entersState(connection, VoiceConnectionStatus.Connecting, 5000),
                ]);
                // Seems to be reconnecting to a new channel - ignore disconnect
            }
            catch (error) {
                // Seems to be a real disconnect which SHOULDN'T be recovered from
                sLogger.log(`Disconnected from ${this.channelName}`, "WARNING");
                this.destruct();
            }
        });
    }
    isSameChannel(channel) {
        return channel?.id == this.voiceChannelID;
    }
    static validateUser(msg) {
        const channel = msg.member?.voice.channel;
        // * This Bot and every others cannot enter *undefined* channel
        if (!channel)
            return VoiceValidateResult.NO_CHANNEL;
        // * This Bot doesn't support Stage Channel
        if (channel instanceof StageChannel)
            return VoiceValidateResult.STAGE_CHANNEL;
        if (!channel.joinable)
            return VoiceValidateResult.NOT_JOINABLE;
        return VoiceValidateResult.OK;
    }
    destruct() {
        const connection = getVoiceConnection(this.guildID);
        connection?.destroy();
        this.handleDisconnect();
    }
    async waitTillReady(mode) {
        const connection = getVoiceConnection(this.guildID);
        if (!connection)
            return false;
        try {
            await entersState(connection, VoiceConnectionStatus.Ready, 5000);
            sLogger.log(`Successfully joined ${this.channelName} as ${chalk.magenta(mode)} 🤩🤩`, "SUCCESS");
            return true;
        }
        catch (err) {
            sLogger.log(`Error joining ${this.channelName} : ${err}`, "ERROR");
            return false;
        }
    }
    async speak(content) {
        sLogger.log(`[TTS] Started Speaking ${shorten(content, 30)} to ${this.channelName}`);
        const results = getAllAudioUrls(content, {
            lang: "th",
            slow: false,
        });
        const initiated = !!this.speakQueue.length;
        this.speakQueue = this.speakQueue.concat(results.map((res) => createAudioResource(res.url, {
            inputType: StreamType.OggOpus,
        })));
        if (!initiated)
            this.player.play(this.speakQueue.shift());
        return new Promise((resolve, reject) => {
            this.player.on(AudioPlayerStatus.Idle, (() => {
                if (this.speakQueue.length) {
                    this.player.play(this.speakQueue.shift());
                }
                else {
                    resolve(true);
                }
            }).bind(this));
            // * By GitHub Copilot
            this.player.on("error", (err) => {
                sLogger.log(`[TTS] Error Speaking ${shorten(content, 30)} to ${this.channelName} : ${err}`, "ERROR");
                resolve(false);
            });
        });
    }
    async playSong(url, title, category) {
        // https://stackoverflow.com/questions/63199238/discord-js-ytdl-error-input-stream-status-code-416
        const musicStream = ytdl(url, {
            filter: "audioonly",
            quality: "highestaudio",
            highWaterMark: 1 << 25,
            liveBuffer: 4000,
        });
        const musicRc = createAudioResource(musicStream);
        sLogger.log(`[DJCorgi] Playing ${title} in category of ${category}`);
        this.player.play(musicRc);
        this.songRunning = true;
        return new Promise((resolve, reject) => {
            this.player.on(AudioPlayerStatus.Idle, (() => {
                this.songRunning = false;
                resolve(true);
            }).bind(this));
            this.player.on("error", (err) => {
                this.songRunning = false;
                sLogger.log(`[DJCorgi] Error while playing ${title}: ${err.message}`, "ERROR");
                resolve(false);
            });
        });
    }
    forceSkip() {
        // * Stop player causing class to think tts/music has ended
        this.player.stop(true);
    }
}
